unit CMSDemoDM;

interface

uses
  Forms, SysUtils, Classes, SecComp, StreamSecII, MpCms, MpX509, ReadStrm,
  Pkix_Cert;

type
  TdmCMSDemo = class(TDataModule)
    MyCertificates: TX509TrustedCertificates;
    SignedData: TMPSignedData;
    PrivateKeyRing: TStreamSecII;
    RootCertificates: TX509TrustedCertificates;
    CollectedCertificates: TX509TrustedCertificates;
    EnvelopedData: TMPEnvelopedData;
    Broker: TMPCMSBroker;
    procedure SignedDataGetDetachedMessage(Sender: TObject;
      var AMessage: TCustomMemoryStream);
    procedure BrokerAddSigner(Sender: TObject; Index: Integer;
      var Cert: TCertificate; var Done: Boolean);
    procedure BrokerAddRecipient(Sender: TObject; Index: Integer;
      var Cert: TCertificate; var Done: Boolean);
    procedure SignedDataGoodSignature(Sender: TObject;
      SignerCert: TCertificate);
    procedure SignedDataGoodCounterSignature(Sender: TObject; SignerCert,
      CounterSignerCert: TCertificate; TimeStamp: TDateTime);
  private
    FPrivateKeyRingFileName: string;
    FCollectedCertsFileName: string;
    FActionDescriptions: TStrings;
  public
    function ProcessData(Actions: TContentTypeArray;
                         Description: string;
                         InStream, OutStream: TStream;
                         IncludeRootCertificates: Boolean): Boolean;
    function CounterSignData(AStream: TStream;
                             Description: string;
                             IncludeRootCertificates: Boolean): Boolean;
    function DecodeData(InStream, OutStream: TStream;
                        Actions: TStrings): Boolean;
    function VerifyData(InStream, OutStream: TStream;
                        Actions: TStrings): Boolean;
    procedure LoadCollectedCerts(AFileName: string);
    procedure LoadMyCerts(AFileName: string);  
    function LoadPrivateKeyRing(AFileName: string): Boolean;
    procedure LoadRootCerts(AFileName: string);
  end;

var
  dmCMSDemo: TdmCMSDemo;

implementation

uses
  Pkix, Asn1, MpPK,
  CMSDemoMain;

{$R *.dfm}

function TdmCMSDemo.DecodeData(InStream, OutStream: TStream;
  Actions: TStrings): Boolean;
var
  Idx: Integer;
  SignCert: PASN1Struct;
  SignKey: IMPPrivateKey;
begin
  Result := Broker.Decode(Instream,OutStream,Actions);
  if Result and CollectedCertificates.Modified then begin
    SignCert := nil;
    Idx := 0;
    if not MyCertificates.FindCert([],[rsaEncryption],[keyCertSign],[],Idx,SignCert) then begin
      Idx := 0;
      if not MyCertificates.FindCert([],[rsaEncryption],[digitalSignature],[],Idx,SignCert) then
        SignCert := nil;
    end;
    if Assigned(SignCert) then begin
      SignKey := PrivateKeyRing.FindCreatePrivateKey(SignCert^);
      if Assigned(SignKey) then
        CollectedCertificates.SignSaveToFile(FCollectedCertsFileName,
                                             SignCert^,
                                             SignKey);
    end;
  end;
end;

procedure TdmCMSDemo.LoadCollectedCerts(AFileName: string);
begin
  CollectedCertificates.LoadFromFile(AFileName);
  FCollectedCertsFileName := AFileName;
end;

procedure TdmCMSDemo.LoadMyCerts(AFileName: string);
begin
  MyCertificates.LoadFromFile(AFileName);
end;

function TdmCMSDemo.LoadPrivateKeyRing(AFileName: string): Boolean;
begin
  if AFileName <> FPrivateKeyRingFileName then begin
    Result := PrivateKeyRing.LoadPrivateKeyRingFromFile(AFileName,nil);
    if Result then
      FPrivateKeyRingFileName := AFileName
    else
      FPrivateKeyRingFileName := '';
  end else
    Result := True;
end;

procedure TdmCMSDemo.LoadRootCerts(AFileName: string);
begin
  RootCertificates.LoadFromFile(AFileName);
end;

function TdmCMSDemo.ProcessData(Actions: TContentTypeArray; Description: string;
  InStream, OutStream: TStream; IncludeRootCertificates: Boolean): Boolean;
begin
  SignedData.SigningDescription := Description;
  SignedData.IncludeRootCertificate := IncludeRootCertificates;
  if not Broker.Encode(Actions,InStream,OutStream) then
    raise Exception.Create('Unable to process data');
  Result := True;
end;

procedure TdmCMSDemo.SignedDataGetDetachedMessage(Sender: TObject;
  var AMessage: TCustomMemoryStream);
begin
  if FrmSign.chbDetachedMessage.Checked then
    begin
      TSecureMemoryStream(AMessage).LoadFromFile(FrmSign.Edit3.Text);
      AMessage.Position := 0;
//      AMessage.WriteBuffer(Content.Memory^, Content.Size);
    end;
end;

procedure TdmCMSDemo.BrokerAddSigner(Sender: TObject; Index: Integer;
  var Cert: TCertificate; var Done: Boolean);                              
var
  Idx: Integer;
  PCert: PASN1Struct;
begin
  PCert := nil;
  if Index = 0 then begin
    Idx := 0;
    if not MyCertificates.FindCert([],[],[digitalSignature,keyEncipherment],[],Idx,PCert) then
      raise Exception.Create('Unable to find a digital signature certificate');
  end;
  if Assigned(PCert) then begin
    Cert.AssignStruct(PCert^);
    Done := False;
  end else
    Done := True;
end;

procedure TdmCMSDemo.BrokerAddRecipient(Sender: TObject; Index: Integer;
  var Cert: TCertificate; var Done: Boolean);
var
  Idx: Integer;
  PCert: PASN1Struct;
begin
  PCert := nil;
  if Index = 0 then begin
    // The first certificate should be your own, to enable you to decrypt the
    // data later:
    Idx := 0;
    if not MyCertificates.FindCert([],[rsaEncryption],[keyEncipherment],[],Idx,PCert) then
      raise Exception.Create('Unable to find a key encipherment certificate');
  end else if Index = 1 then begin
    // The following certificates are the certificates of the recipients.
    // You MUST add some checks here to make sure that the content is encrypted
    // for the intended recipients only. Here, all if any certificates with
    // rsaEncryption keyEncipherment public keys present in
    // CollectedCertificates are returned:
    Idx := 0;
    if not CollectedCertificates.FindCert([],[rsaEncryption],[keyEncipherment],[],Idx,PCert) then
      PCert := nil;
  end;
  if Assigned(PCert) then begin
    if Cert.Data.Length > 0 then begin
      // Destroying the old certificate is more efficient:
      Cert.Free;
      Cert := TCertificate.Create(nil,nil);
    end;
    Cert.AssignStruct(PCert^);
    Done := False;
  end else
    Done := True;
end;

function TdmCMSDemo.VerifyData(InStream, OutStream: TStream;
  Actions: TStrings): Boolean;
var
  Idx: Integer;
  SignCert: PASN1Struct;
  SignKey: IMPPrivateKey;
begin
  FActionDescriptions := Actions;
  try
    Result := SignedData.VerifySignedData(Instream,OutStream);
    if Result and CollectedCertificates.Modified then begin
      SignCert := nil;
      Idx := 0;
      if not MyCertificates.FindCert([],[rsaEncryption],[keyCertSign],[],Idx,SignCert) then begin
        Idx := 0;
        if not MyCertificates.FindCert([],[rsaEncryption],[digitalSignature],[],Idx,SignCert) then
          SignCert := nil;
      end;
      if Assigned(SignCert) then begin
        SignKey := PrivateKeyRing.FindCreatePrivateKey(SignCert^);
        if Assigned(SignKey) then
          CollectedCertificates.SignSaveToFile(FCollectedCertsFileName,
                                               SignCert^,
                                               SignKey);
      end;
    end;
  finally
    FActionDescriptions := nil;
  end;
end;

procedure TdmCMSDemo.SignedDataGoodSignature(Sender: TObject;
  SignerCert: TCertificate);
begin
  if Assigned(FActionDescriptions) then begin
    FActionDescriptions.Append('SignedData');
    if SignedData.SigningDescription <> '' then
      FActionDescriptions.Append('  Description: ' + SignedData.SigningDescription);
    with SignerCert.Subject do begin
      LineSeparator := '/';
      if CommonName <> '' then
        FActionDescriptions.Append('  CN=' + CommonName);
      if Title <> '' then
        FActionDescriptions.Append('  T=' + Title);
      if GivenName <> '' then
        FActionDescriptions.Append('  GN=' + GivenName);
      if Initials <> '' then
        FActionDescriptions.Append('  I=' + Initials);
      if Surname <> '' then
        FActionDescriptions.Append('  S=' + Surname);
      if GenerationQualifier <> '' then
        FActionDescriptions.Append('  GQ=' + GenerationQualifier);
      if OrganizationName <> '' then
        FActionDescriptions.Append('  O=' + OrganizationName);
      if OrganizationalUnitName <> '' then
        FActionDescriptions.Append('  CN=' + OrganizationalUnitName);
      if LocalityName <> '' then
        FActionDescriptions.Append('  L=' + LocalityName);
      if StateOrProvinceName <> '' then
        FActionDescriptions.Append('  S=' + StateOrProvinceName);
      if Country <> '' then
        FActionDescriptions.Append('  C=' + Country);
      if DnQualifier <> '' then
        FActionDescriptions.Append('  DQ=' + DnQualifier);
    end;
  end;
end;

function TdmCMSDemo.CounterSignData(AStream: TStream; Description: string;
  IncludeRootCertificates: Boolean): Boolean;
begin
  SignedData.SigningDescription := Description;
  SignedData.IncludeRootCertificate := IncludeRootCertificates;
  if not Self.SignedData.CounterSignData(nil,AStream,AStream) then
    raise Exception.Create('Unable to counter sign file');
  Result := True;
end;

procedure TdmCMSDemo.SignedDataGoodCounterSignature(Sender: TObject;
  SignerCert, CounterSignerCert: TCertificate; TimeStamp: TDateTime);
begin
  if Assigned(FActionDescriptions) then begin
    FActionDescriptions.Append('SignedData Counter Signature');
    if SignedData.SigningDescription <> '' then
      FActionDescriptions.Append('  Description: ' + SignedData.SigningDescription);
    if TimeStamp <> 0 then
      FActionDescriptions.Append('  Signing Time: ' + DateTimeToStr(TimeStamp));
    with CounterSignerCert.Subject do begin
      LineSeparator := '/';
      if CommonName <> '' then
        FActionDescriptions.Append('  CN=' + CommonName);
      if Title <> '' then
        FActionDescriptions.Append('  T=' + Title);
      if GivenName <> '' then
        FActionDescriptions.Append('  GN=' + GivenName);
      if Initials <> '' then
        FActionDescriptions.Append('  I=' + Initials);
      if Surname <> '' then
        FActionDescriptions.Append('  S=' + Surname);
      if GenerationQualifier <> '' then
        FActionDescriptions.Append('  GQ=' + GenerationQualifier);
      if OrganizationName <> '' then
        FActionDescriptions.Append('  O=' + OrganizationName);
      if OrganizationalUnitName <> '' then
        FActionDescriptions.Append('  CN=' + OrganizationalUnitName);
      if LocalityName <> '' then
        FActionDescriptions.Append('  L=' + LocalityName);
      if StateOrProvinceName <> '' then
        FActionDescriptions.Append('  S=' + StateOrProvinceName);
      if Country <> '' then
        FActionDescriptions.Append('  C=' + Country);
      if DnQualifier <> '' then
        FActionDescriptions.Append('  DQ=' + DnQualifier);
    end;
  end;
end;

end.
