unit SignedDataDM;

interface

uses
  SysUtils, Classes, Forms,
  Asn1, MpX509, MpCms, Pkix_Cert, SecComp, StreamSecII;

type
  TdmSignedData = class(TDataModule)
    MPSignedData1: TMPSignedData;
    CollectedCerts: TX509TrustedCertificates;
    RootCerts: TX509TrustedCertificates;
    PrivateKeyRing: TStreamSecII;
    MyCerts: TX509TrustedCertificates;
    procedure DataModuleCreate(Sender: TObject);
    procedure CollectedCertsCertNotAccepted(Sender: TObject;
      Cert: TASN1Struct; Status: TCertStatusCode);
    procedure CollectedCertsCertNotTrusted(Sender: TObject;
      Cert: TASN1Struct; var ExplicitTrust: Boolean);
    procedure MPSignedData1BadSignature(Sender: TObject;
      SignerCert: TCertificate);
    procedure MPSignedData1GoodSignature(Sender: TObject;
      SignerCert: TCertificate);
  private
    FGoodSignatures: Integer;
    FBadSignatures: Integer;
    FBadCertificates: Integer;
    procedure LoadDefault;
    procedure SetBadSignatures(const Value: Integer);
    procedure SetGoodSignatures(const Value: Integer);
    procedure SetBadCertificates(const Value: Integer);
  public
    function Verify(const SrcFileName, DstFileName: string): Boolean;
    property BadCertificates: Integer read FBadCertificates write SetBadCertificates;
    property BadSignatures: Integer read FBadSignatures write SetBadSignatures;
    property GoodSignatures: Integer read FGoodSignatures write SetGoodSignatures;
  end;

var
  dmSignedData: TdmSignedData;

implementation

uses
  X509Base, Controls, CertViewDlg;

{$R *.dfm}

{ TdmSignedData }

procedure TdmSignedData.LoadDefault;
var
  Dir: string;
begin
  Dir := ExtractFilePath(ParamStr(0));
  Delete(Dir,Length(Dir),1);
  Dir := ExtractFilePath(Dir) + 'TLS\';
  RootCerts.LoadFromFile(Dir + 'Root.scl');
  { TODO : Load a private signing key and a corresponding certificate }
end;

procedure TdmSignedData.DataModuleCreate(Sender: TObject);
begin
  LoadDefault;
end;

procedure TdmSignedData.CollectedCertsCertNotAccepted(Sender: TObject;
  Cert: TASN1Struct; Status: TCertStatusCode);
var
  T: TCertificate;
  RC: TX509RevokedCertificate;
begin
  T := TCertificate.Create(nil,nil);
  try
    T.AssignStruct(Cert);
    RC := nil;
    try
      if not RootCerts.FindRevocation(T.Data,RC) then
        CollectedCerts.FindRevocation(T.Data,RC);
      case Status of
        crcExpired:              frmCertificateViewer.Caption := 'Expired Certificate';
        crcInvalidSignature:     frmCertificateViewer.Caption := 'Invalid Signature';
        crcCANotTrusted:         frmCertificateViewer.Caption := 'CA Not Trusted';
        crcRevoked:              frmCertificateViewer.Caption := 'Revoked Certificate';
        crcCAExpiredOrRevoked:   frmCertificateViewer.Caption := 'CA Expired or Revoked';
        crcConstraintFault:      frmCertificateViewer.Caption := 'Constraint Fault';
        crcInvalidKeyUsage:      frmCertificateViewer.Caption := 'Invalid Key Usage';
        crcUnsupportedExtension: frmCertificateViewer.Caption := 'Unsupported Extension';
        crcSuperseeded:          frmCertificateViewer.Caption := 'Superseeded';
        crcSyntax:               frmCertificateViewer.Caption := 'Syntax Error';
        crcBaseNotFound:         frmCertificateViewer.Caption := 'Base Not Found';
        crcTrusted:              frmCertificateViewer.Caption := 'Trusted';
        crcTooSmallKey:          frmCertificateViewer.Caption := 'Too Small Key';
        crcPolicyNotAccepted:    frmCertificateViewer.Caption := 'Policy Not Accepted';
      else
        frmCertificateViewer.Caption := 'Certificate not accepted';
      end;
      frmCertificateViewer.CertificateViewer1.Display(T,RC,nil,nil);
      frmCertificateViewer.ShowModal;
      Inc(FBadCertificates);
    finally
      RC.Free;
    end;
  finally
    T.Free;
  end;
  frmCertificateViewer.Caption := 'Certificate';
end;

procedure TdmSignedData.CollectedCertsCertNotTrusted(Sender: TObject;
  Cert: TASN1Struct; var ExplicitTrust: Boolean);
var
  T: TCertificate;
  RC: TX509RevokedCertificate;
begin
  T := TCertificate.Create(nil,nil);
  try
    T.AssignStruct(Cert);
    RC := nil;
    try
      if not RootCerts.FindRevocation(T.Data,RC) then
        CollectedCerts.FindRevocation(T.Data,RC);
      frmCertificateViewer.Caption := 'Certificate not trusted';
      frmCertificateViewer.btnAccept.Visible := True;
      try
        frmCertificateViewer.CertificateViewer1.Display(T,RC,nil,nil);
        ExplicitTrust := (frmCertificateViewer.ShowModal = mrRetry);
        if not ExplicitTrust then
          Inc(FBadCertificates);
      finally
        frmCertificateViewer.btnAccept.Visible := False;
      end;
    finally
      RC.Free;
    end;
  finally
    T.Free;
  end;
  frmCertificateViewer.Caption := 'Certificate';
end;

procedure TdmSignedData.MPSignedData1BadSignature(Sender: TObject;
  SignerCert: TCertificate);
var
  RC: TX509RevokedCertificate;
begin
  RC := nil;
  try
    if not RootCerts.FindRevocation(SignerCert.Data,RC) then
      CollectedCerts.FindRevocation(SignerCert.Data,RC);
    frmCertificateViewer.Caption := 'Bad signature';
    frmCertificateViewer.CertificateViewer1.Display(SignerCert,RC,nil,nil);
    frmCertificateViewer.ShowModal;
  finally
    RC.Free;
  end;
  frmCertificateViewer.Caption := 'Certificate';
  Inc(FBadSignatures);
end;

procedure TdmSignedData.SetBadSignatures(const Value: Integer);
begin
  FBadSignatures := Value;
end;

procedure TdmSignedData.SetGoodSignatures(const Value: Integer);
begin
  FGoodSignatures := Value;
end;

function TdmSignedData.Verify(const SrcFileName,
  DstFileName: string): Boolean;
var
  Src, Dst: TFileStream;
begin
  FGoodSignatures := 0;
  FBadSignatures := 0;
  FBadCertificates := 0;
  Src := TFileStream.Create(SrcFileName,fmOpenRead);
  try
    Dst := TFileStream.Create(DstFileName,fmCreate);
    try
      Result := MPSignedData1.VerifySignedData(Src,Dst);
    finally
      Dst.Free;
    end;
  finally
    Src.Free;
  end;
end;

procedure TdmSignedData.SetBadCertificates(const Value: Integer);
begin
  FBadCertificates := Value;
end;

procedure TdmSignedData.MPSignedData1GoodSignature(Sender: TObject;
  SignerCert: TCertificate);
var
  RC: TX509RevokedCertificate;
begin
  RC := nil;
  try
    if not RootCerts.FindRevocation(SignerCert.Data,RC) then
      CollectedCerts.FindRevocation(SignerCert.Data,RC);
    frmCertificateViewer.Caption := 'Good signature';
    frmCertificateViewer.CertificateViewer1.Display(SignerCert,RC,nil,nil);
    frmCertificateViewer.ShowModal;
  finally
    RC.Free;
  end;
  frmCertificateViewer.Caption := 'Certificate';
  Inc(FGoodSignatures);
end;

end.
