unit CertMgrCliModule;

interface

uses
  SysUtils, Classes, uROServiceComponent, uRORemoteService, uROClient,
  uROTypes, uROBINMessage, uROBaseConnection, uROProxy, uRONamedPipeChannel;

type
  TClientModule = class(TDataModule)
    ROChannel: TRONamedPipeChannel;
    ROMessage: TROBinMessage;
    RORemoteService: TRORemoteService;
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
  private
    FHandle: LongInt;
    FCACerts: TMemoryStream;
    FCertReq: Binary;
  public
    procedure ListCACerts(out CACerts: Pointer; out CACertsLen: LongInt);
  end;

function CreateClient: LongInt; stdcall;
procedure FreeClient(AHandle: LongInt); stdcall;

function LogIn(AHandle: LongInt;
               AStoreName, AStorePath: PWideChar): Boolean; stdcall;
procedure LogOut(AHandle: LongInt); stdcall;

procedure AddCert(AHandle: LongInt;
                  Cert: Pointer; CertLen: LongInt); stdcall;
procedure RemoveCert(AHandle: Integer;
                     Cert: Pointer; CertLen: LongInt); stdcall;
function FindCert(AHandle: LongInt;
                  Cert: Pointer; CertLen: LongInt): Boolean; stdcall;
function ActivateQuery(AHandle: Integer): Integer; stdcall;
procedure ListCACerts(AHandle: LongInt;
                      out CACerts: Pointer; out CACertsLen: LongInt); stdcall;
procedure SignCertReq(AHandle: LongInt;
                      ACACertStoreName: PWideChar;
                      CertReq: Pointer; CertReqLen: LongInt;
                      Cert: Pointer; var CertLen: LongInt); stdcall;
function CheckSignCertReq(AHandle: LongInt;
                          out CertReq: Pointer; out CertReqLen: LongInt): Boolean; stdcall;
procedure SignedCertReq(AHandle: LongInt;
                        Cert: Pointer; CertLen: LongInt); stdcall;

implementation

uses
  Windows, ShellAPI,
  SsCertMgrUtils,
  CertMgrSrv_Intf;

{$R *.dfm}

function CreateClient: LongInt; stdcall;
var
  DM: TClientModule;
  ExeName: string;
  Retries: Cardinal;
begin
  Result := 0;
  DM := TClientModule.Create(nil);
  try
    try
      DM.ROChannel.Active := True;
    except
      ExeName := TSsCertMgrPaths.GetCertMgrSrvExeName;
      if ExeName = '' then begin
        SetLength(ExeName,1024);
        SetLength(ExeName,GetModuleFileName(HINSTANCE,PChar(ExeName),Length(ExeName)));
        ExeName := Format('%sCertMgrSrv.exe',[ExtractFilePath(ExeName)]);
        if FileExists(ExeName) then
          TSsCertMgrPaths.RegisterCertMgrSrv(ExeName)
        else
          raise;
      end;
      ShellExecute(0,'OPEN',PChar(ExeName),'','',SW_SHOWNORMAL);
      Retries := 10;
      while Retries > 0 do begin
        Sleep(1000);
        try
          DM.ROChannel.Active := True;
          Retries := 0;
        except
          DM.ROChannel.Active := False;
          Dec(Retries);
        end;
      end;
    end;
    Result := LongInt(DM);
  except
    DM.Free;
  end;
end;

procedure FreeClient(AHandle: LongInt); stdcall;
begin
  with TClientModule(AHandle) do try
    if ROChannel.Active then
      ROChannel.Disconnect;
  finally
    Free;
  end;
end;

function LogIn(AHandle: LongInt; AStoreName, AStorePath: PWideChar): Boolean;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
begin                                        
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    DM.FHandle := Intf.SetStore(AStoreName,AStorePath);
    Result := True;
  except
    Result := False;
  end;
end;

procedure LogOut(AHandle: LongInt); stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Intf.CloseStore(DM.FHandle);
  except
  end;
end;

procedure AddCert(AHandle: Integer;
                  Cert: Pointer; CertLen: LongInt); stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
  Crt: Binary;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Crt := Binary.Create;
    try
      Crt.Write(Cert^,CertLen);
      Crt.Position := 0;
      Intf.AddCert(DM.FHandle,Crt);
    finally
      Crt.Free;
    end;
  except
  end;
end;

procedure RemoveCert(AHandle: Integer;
                     Cert: Pointer; CertLen: LongInt); stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
  Crt: Binary;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Crt := Binary.Create;
    try
      Crt.Write(Cert^,CertLen);
      Crt.Position := 0;
      Intf.RemoveCert(DM.FHandle,Crt);
    finally
      Crt.Free;
    end;
  except
  end;
end;

function FindCert(AHandle: LongInt; Cert: Pointer; CertLen: LongInt): Boolean; stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
  Crt: Binary;
begin
  DM := TClientModule(AHandle);
  Intf := DM.RORemoteService as ICertMgrSrvc;
  Crt := Binary.Create;
  try
    Crt.Write(Cert^,CertLen);
    Crt.Position := 0;
    Result := Intf.FindCert(Crt);
  finally
    Crt.Free;
  end;
end;

function ActivateQuery(AHandle: Integer): Integer; stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Result := Intf.ActivateQuery(DM.FHandle);
  except
    Result := -1;
  end;
end;

procedure ListCACerts(AHandle: LongInt;
                      out CACerts: Pointer; out CACertsLen: LongInt); stdcall;
var
  DM: TClientModule;
begin
  DM := TClientModule(AHandle);
  DM.ListCACerts(CACerts,CACertsLen);
end;

procedure SignCertReq(AHandle: LongInt;
                      ACACertStoreName: PWideChar;
                      CertReq: Pointer; CertReqLen: LongInt;
                      Cert: Pointer; var CertLen: LongInt); stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
  CrtReq, Crt: Binary;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    CrtReq := Binary.Create;
    try
      CrtReq.Write(CertReq^,CertReqLen);
      CrtReq.Position := 0;
      Crt := Intf.SignCertReq(ACACertStoreName,CrtReq);
      try
        if CertLen >= Crt.Size then begin
          Crt.Position := 0;
          CertLen := Crt.Read(Cert^,CertLen);
        end;
      finally
        Crt.Free;
      end;
    finally
      CrtReq.Free;
    end;
  except
    CertLen := 0;
  end;
end;

function CheckSignCertReq(AHandle: Integer;
                          out CertReq: Pointer; out CertReqLen: LongInt): Boolean; stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
begin
  try
    DM := TClientModule(AHandle);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Result := Intf.CheckSignCertReq(DM.FHandle,DM.FCertReq);
    if Result then begin
      CertReq := DM.FCertReq.Memory;
      CertReqLen := DM.FCertReq.Size;
    end;
  except
    Result := False;
  end;
end;

procedure SignedCertReq(AHandle: Integer;
                        Cert: Pointer; CertLen: LongInt); stdcall;
var
  DM: TClientModule;
  Intf: ICertMgrSrvc;
  Crt: Binary;
begin
  try
    DM := TClientModule(AHandle);
    FreeAndNil(DM.FCertReq);
    Intf := DM.RORemoteService as ICertMgrSrvc;
    Crt := Binary.Create;
    try
      Crt.Write(Cert^,CertLen);
      Crt.Position := 0;
      Intf.SignedCertReq(DM.FHandle,Crt);
    finally
      Crt.Free;
    end;
  except
  end;
end;

procedure TClientModule.DataModuleCreate(Sender: TObject);
begin
  FHandle := -1;
end;

procedure TClientModule.DataModuleDestroy(Sender: TObject);
begin
  FCACerts.Free;
end;

procedure TClientModule.ListCACerts(out CACerts: Pointer;
  out CACertsLen: Integer);
var
  Intf: ICertMgrSrvc;
begin
  if FCACerts = nil then begin
    Intf := RORemoteService as ICertMgrSrvc;
    FCACerts := Intf.ListCACerts;
  end;
  CACerts := FCACerts.Memory;
  CACertsLen := FCACerts.Size;
end;

end.
