unit CertStoreModule;

interface

uses
  SysUtils, Classes, SyncObjs, MpX509, uROTypes, ExtCtrls;

type
  TStoreModule = class(TDataModule)
    Certs: TX509TrustedCertificates;
    procedure DataModuleCreate(Sender: TObject);
    procedure LaunchTimerTimer(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
  private
    FLock: TCriticalSection;
    FStorePath: WideString;
    FStoreName: WideString;
    FIndex: Integer;
    FSelected: Integer;
    FCertReqs: TThreadList;
    procedure SetSelected(const Value: Integer);
  public
    procedure Lock;
    procedure UnLock;
    procedure ActivateCert(Cert: Binary);  
    procedure ActivateAny;
    procedure AddCert(Cert: Binary);
    function FindCert(Cert: Binary): Boolean;
    procedure ListCACerts(CertList: Binary);
    function LogIn: Integer;
    procedure LogOut;
    procedure RemoveCert(Cert: Binary);
    function SignCertReq(CertReq: Binary): Binary;
    function CheckSignCertReq(out CertReq: Binary): Boolean;
    procedure SignedCertReq(const Cert: Binary);
    property Index: Integer read FIndex;
    property Selected: Integer read FSelected write SetSelected;
    property StorePath: WideString read FStorePath;
    property StoreName: WideString read FStoreName;
  end;

function FindCert(Cert: Binary): TStoreModule;
function FindModule(Index: Integer): TStoreModule; overload;
function FindModule(const StoreName, StorePath: WideString): TStoreModule; overload;
procedure ListCACerts(CertList: Binary);

implementation

uses
  Asn1, X509Base, SsCertMgrUtils,
  Windows, ShellAPI;

{$R *.dfm}

type
  TSignReq = class
  private
    FList: TThreadList;
    FCertReq: Binary;
    FCert: Binary;
  public
    constructor Create(AList: TThreadList; ACertReq: Binary);
    destructor Destroy; override;
  end;

var
  GStores: TThreadList;
  GNextIndex: Integer;

function CreateModule(List: TList; const StoreName, StorePath: WideString): TStoreModule;
var
  FN: string;
begin
  FN := StorePath + 'User.scl';
  Result := TStoreModule.Create(nil);
  try
    if FileExists(FN) then
      Result.Certs.LoadFromFile(FN);
    Result.FStorePath := StorePath;
    Result.FStoreName := StoreName;
    Result.FIndex := GNextIndex;
    Inc(GNextIndex);
    List.Add(Result);
  except
    Result.Free;
    Result := nil;
  end;
end;

procedure CreateAllModules;
var
  SL: TStringList;
  I: Integer;
  List: TList;
begin
  GStores := TThreadList.Create;
  SL := TStringList.Create;
  try
    TSsCertMgrPaths.GetPaths(SL);
    List := GStores.LockList;
    try
      for I := 0 to SL.Count - 1 do
        CreateModule(List,TSsCertMgrPaths.GetPathName(SL[I]),SL[I]);
    finally
      GStores.UnlockList;
    end;
  finally
    SL.Free;
  end;
end;

procedure FreeAllModules;
var
  List: TList;
  I: Integer;
  DM: TStoreModule;
begin
  List := GStores.LockList;
  try
    for I := List.Count - 1 downto 0 do begin
      DM := List[I];
      List.Delete(I);
      DM.FreeOnRelease;
    end;
  finally
    GStores.UnlockList;
  end;
  GStores.Free;
end;

function FindModule(const StoreName, StorePath: WideString): TStoreModule;
var
  List: TList;
  I: Integer;
begin
  List := GStores.LockList;
  try
    Result := nil;
    for I := List.Count - 1 downto 0 do begin
      Result := List[I];
      if ((Result.StorePath = StorePath) or (StorePath = '')) and
         (Result.StoreName = StoreName) then
        Break
      else
        Result := nil;
    end;
    if Result = nil then
      Result := CreateModule(List,StoreName,StorePath);
  finally
    GStores.UnlockList;
  end;
end;

function FindModule(Index: Integer): TStoreModule;
var
  List: TList;
  L, H, I, C: Integer;
begin
  Result := nil;
  List := GStores.LockList;
  try
    L := 0;
    H := List.Count - 1;
    if (Index >= L) and (Index <= H) and (TStoreModule(List[Index]).Index = Index) then
      Result := List[Index]
    else begin
      while L <= H do begin
        I := (L + H) shr 1;
        C := TStoreModule(List[I]).Index - Index;
        if C < 0 then
          L := I + 1
        else begin
          H := I - 1;
          if C = 0 then begin
            Result := List[I];
            L := I;
          end;
        end;
      end;
    end;
  finally
    GStores.UnlockList;
  end;
end;

function FindCert(Cert: Binary): TStoreModule;
var
  List: TList;
  I: Integer;
begin
  List := GStores.LockList;
  try
    for I := List.Count - 1 downto 0 do begin
      Result := List[I];
      if Result.FindCert(Cert) then
        Break
      else
        Result := nil;
    end;
  finally
    GStores.UnlockList;
  end;
end;

procedure ListCACerts(CertList: Binary);      
var
  List: TList;
  I: Integer;
begin
  List := GStores.LockList;
  try
    for I := 0 to List.Count - 1 do
      TStoreModule(List[I]).ListCACerts(CertList);
  finally
    GStores.UnlockList;
  end;
end;

{ TStoreModule }

procedure TStoreModule.ActivateCert(Cert: Binary);
var
  F: TASN1Struct;
begin
  Lock;
  try
    if Cert = nil then
      FSelected := -1
    else begin
      F := TASN1Struct.Create;
      try
        Cert.Position := 0;
        F.LoadFromStream(Cert,fmtDER);
        FSelected := Certs.IndexOfCert(F);
        if FSelected > 0 then
          if not TSsCertMgrPaths.PathIsLocked(FStorePath) then
            LaunchTimerTimer(nil);
      finally
        F.Free;
      end;
    end;
  finally
    UnLock;
  end;
end;

procedure TStoreModule.AddCert(Cert: Binary);
var
  F: TASN1Struct;
  Status: TCertStatusCode;
begin
  Lock;
  try
    F := TASN1Struct.Create;
    try
      Cert.Position := 0;
      F.LoadFromStream(Cert,fmtDER);
      Certs.AddCertificate(F,True,Status);
    finally
      F.Free;
    end;
  finally
    UnLock;
  end;
end;

function TStoreModule.FindCert(Cert: Binary): Boolean;
var
  F: TASN1Struct;
begin
  Lock;
  try
    F := TASN1Struct.Create;
    try
      Cert.Position := 0;
      F.LoadFromStream(Cert,fmtDER);
      Result := Certs.IndexOfCert(F) >= 0;
    finally
      F.Free;
    end;
  finally
    UnLock;
  end;
end;

procedure TStoreModule.ListCACerts(CertList: Binary);
var
  I: Integer;
begin
  Lock;
  try
    for I := 0 to Certs.Count - 1 do
      if X509Base.ExtractPathLenConstraint(Certs.Certs[I]) >= 0 then
        Certs.Certs[I].SaveToStream(CertList,fmtDER);
  finally
    UnLock;
  end;
end;

procedure TStoreModule.Lock;
begin
  FLock.Acquire;
end;

procedure TStoreModule.RemoveCert(Cert: Binary);
var
  F: TASN1Struct;
  Status: TCertStatusCode;
begin
  Lock;
  try
    F := TASN1Struct.Create;
    try
      Cert.Position := 0;
      F.LoadFromStream(Cert,fmtDER);
      Certs.RemoveCertificate(F);
    finally
      F.Free;
    end;
  finally
    UnLock;
  end;
end;

procedure TStoreModule.SetSelected(const Value: Integer);
begin
  FSelected := Value;
end;

procedure TStoreModule.UnLock;
begin
  FLock.Release;
end;

procedure TStoreModule.DataModuleCreate(Sender: TObject);
begin
  FLock := TCriticalSection.Create;
  FCertReqs := TThreadList.Create;
  FSelected := -1;
end;

procedure TStoreModule.Logout;       
var
  FN: string;
begin
  FN := FStorePath + 'User.scl';
  Lock;
  try      
    if FileExists(FN) then
      Certs.LoadFromFile(FN);
    TSsCertMgrPaths.UnLockPath(FStorePath);
  finally
    UnLock;
  end;
end;

procedure TStoreModule.LaunchTimerTimer(Sender: TObject);
var
  ExeName: string;
  Param: string;
begin
  ExeName := TSsCertMgrPaths.GetCertMgrExeName;
  if ExeName = '' then begin
    ExeName := Format('%sCertMgr.exe',[ExtractFilePath(ParamStr(0))]);
    if FileExists(ExeName) then
      TSsCertMgrPaths.RegisterCertMgr(ExeName)
    else
      Exit;
  end;
  Param := Format('"%sUser.pkr" Dummy'#0,[FStorePath]);
  ShellExecute(0,'OPEN',PChar(ExeName),PChar(Param),'',SW_SHOWNORMAL);
end;

function TStoreModule.LogIn: Integer;     
var
  FN: string;
begin
  FN := FStorePath + 'User.scl';
  Lock;
  try
    if TSsCertMgrPaths.PathIsLocked(FStorePath) then
      Result := -1
    else begin
      if FileExists(FN) then
        Certs.LoadFromFile(FN);
      TSsCertMgrPaths.LockPath(FStorePath);
      Result := Index;
    end;
  finally
    UnLock;
  end;
end;

function TStoreModule.SignCertReq(CertReq: Binary): Binary;
var
  SignReq: TSignReq;
begin
  SignReq := TSignReq.Create(FCertReqs,CertReq); 
  if not TSsCertMgrPaths.PathIsLocked(FStorePath) then
    LaunchTimerTimer(nil);
  repeat
    Sleep(50);
  until Assigned(SignReq.FCert);
  Result := SignReq.FCert;
  SignReq.Free;
end;

procedure TStoreModule.DataModuleDestroy(Sender: TObject);
var
  List: TList;
  I: Integer;
begin
  List := FCertReqs.LockList;
  try
    for I := List.Count - 1 downto 0 do
      with TSignReq(List[I]) do try
        FList := nil;
      finally
        Free;
      end;
  finally
    FCertReqs.UnlockList;
  end;
  FCertReqs.Free;
  FLock.Free;
end;

function TStoreModule.CheckSignCertReq(out CertReq: Binary): Boolean;
var
  List: TList;
  I: Integer;
begin
  List := FCertReqs.LockList;
  try
    Result := False;
    for I := 0 to List.Count - 1 do begin
      Result := TSignReq(List[I]).FCert = nil;
      if Result then begin
        CertReq := Binary.Create;
        CertReq.CopyFrom(TSignReq(List[I]).FCertReq,0);
        Break;
      end;
    end;
  finally
    FCertReqs.UnlockList;
  end;
end;

procedure TStoreModule.SignedCertReq(const Cert: Binary);
var
  List: TList;
  I: Integer;
begin
  List := FCertReqs.LockList;
  try
    for I := 0 to List.Count - 1 do
      if TSignReq(List[I]).FCert = nil then begin
        TSignReq(List[I]).FCert := Binary.Create;
        TSignReq(List[I]).FCert.CopyFrom(Cert,0);
        Break;
      end;
  finally
    FCertReqs.UnlockList;
  end;
end;

procedure TStoreModule.ActivateAny;
var
  F: TASN1Struct;
begin
  Lock;
  try
    FSelected := 0;
    if not TSsCertMgrPaths.PathIsLocked(FStorePath) then
      LaunchTimerTimer(nil);
  finally
    UnLock;
  end;
end;

{ TSignReq }

constructor TSignReq.Create(AList: TThreadList; ACertReq: Binary);
begin
  FList := AList;
  FCertReq := ACertReq;
  AList.Add(Self);
end;

destructor TSignReq.Destroy;
begin
  if Assigned(FList) then
    FList.Remove(Self);
  inherited;
end;

initialization
  CreateAllModules;
finalization
  FreeAllModules;
end.
