{$ASSERTIONS ON}
unit DigSignModule;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  MpYarrow, StreamSecII, MpX509, SecUtils, SecComp;

type
  TdmDigSign = class(TDataModule)
    MPYarrow1: TMPYarrow;
    X509TrustedCertificates1: TX509TrustedCertificates;
    StreamSecII1: TStreamSecII;
    procedure DataModuleCreate(Sender: TObject);
    procedure MPYarrow1Reseeded(Sender: TObject; Slow: Boolean);
    procedure MPYarrow1Reseeding(Sender: TObject; Slow: Boolean);
    procedure DataModuleDestroy(Sender: TObject);
    procedure StreamSecII1AdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure StreamSecII1Password(Sender: TObject; Password: ISecretKey);
  private
    FSlowReseeds: Cardinal;
    FFastReseeds: Cardinal;
    FReseedThreads: Cardinal;
    FOnReseed: TNotifyEvent;
    FOnReseeding: TNotifyEvent;
    procedure ApplicationMessage(var Msg: TMsg; var Handled: Boolean);
    procedure SetOnReseed(const Value: TNotifyEvent);
    procedure SetOnReseeding(const Value: TNotifyEvent);
    { Private declarations }
  public
    procedure CreateDLKeyPair(KeySize: Cardinal; NRKey: Boolean; HA: THashAlgorithm);
    procedure CreateECKeyPair(Curve: string; NRKey: Boolean; HA: THashAlgorithm);
    procedure CreateIFKeyPair(KeySize: Cardinal; PSSKey: Boolean; HA: THashAlgorithm);
    procedure ForceSlowReseed(const Buf; Count: Integer);
    function HasDLKeyPair(NRKey: Boolean): Boolean;
    function HasECKeyPair(NRKey: Boolean): Boolean;
    function HasIFKeyPair(PSSKey: Boolean): Boolean;
    function SignMsg(const Msg: string; HA: THashAlgorithm): string;
    function VerifySign(const Msg, Sign: string; HA: THashAlgorithm): Boolean;
    property FastReseeds: Cardinal read FFastReseeds;
    property SlowReseeds: Cardinal read FSlowReseeds;
    property ReseedThreads: Cardinal read FReseedThreads;
    property OnReseed: TNotifyEvent read FOnReseed write SetOnReseed;
    property OnReseeding: TNotifyEvent read FOnReseeding write SetOnReseeding;
  end;

var
  dmDigSign: TdmDigSign;

implementation

uses
  Asn1, MpDL, MpEC, MpIF, X509Base, Pkix, Pkix_Cert, SsRijndael;

{$R *.DFM}

procedure TdmDigSign.ApplicationMessage(var Msg: TMsg;
  var Handled: Boolean);
begin
  MPYarrow1.Accumulate(Msg,SizeOf(Msg),1,0);
end;

procedure TdmDigSign.DataModuleCreate(Sender: TObject);
begin
  if Assigned(dmDigSign) and (dmDigSign <> Self) then
    raise Exception.Create('There can be only one instance of TdmDigSign per application');
  Application.OnMessage := ApplicationMessage;
  MPYarrow1.ForceSlowReseed(nil^,0,1000);
end;

procedure TdmDigSign.SetOnReseed(const Value: TNotifyEvent);
begin
  FOnReseed := Value;
end;

procedure TdmDigSign.MPYarrow1Reseeded(Sender: TObject; Slow: Boolean);
begin
  if FReseedThreads > 0 then
    Dec(FReseedThreads);
  if Slow then
    Inc(FSlowReseeds)
  else
    Inc(FFastReseeds);
  if Assigned(FOnReseed) then
    FOnReseed(Self);
end;

procedure TdmDigSign.ForceSlowReseed(const Buf; Count: Integer);
begin
  MPYarrow1.Accumulate(Buf,Count,0,Count shr 1);
  MPYarrow1.ForceSlowReseed(Buf,Count,500);
end;

procedure TdmDigSign.MPYarrow1Reseeding(Sender: TObject; Slow: Boolean);
begin
  Inc(FReseedThreads);
  if Assigned(FonReseeding) then
    FOnReseeding(Self);
end;

procedure TdmDigSign.SetOnReseeding(const Value: TNotifyEvent);
begin
  FOnReseeding := Value;
end;

procedure TdmDigSign.DataModuleDestroy(Sender: TObject);
var
  MS: TMemoryStream;
begin
// --- This doesn't do anything.
  MS := TMemoryStream.Create;
  try
    StreamSecII1.SavePrivateKeyRingToStream(MS,'abc',kdfWPv2SHA256,1,True,id_aes256_wrap,TRijndael_ABC);
    MS.Position := 0;
    StreamSecII1.LoadPrivateKeyRingFromStream(MS,'abc');
  finally
    MS.Free;
  end;
// ---
  dmDigSign := nil;
end;

procedure TdmDigSign.CreateDLKeyPair;
var
  Cert: TCertificate;
  OID: ObjectIdentifier;
  Status: TCertStatusCode;
begin
  Cert := TCertificate.Create(nil,nil);
  try
    if NRKey then begin
      Cert.Subject.CommonName := 'NR Demo Key';
      OID := id_nr;
    end else begin
      Cert.Subject.CommonName := 'DSA Demo Key';
      OID := id_dsa;
    end;
    Cert.Validity.NotBefore.Choice := teUtcTime;
    Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now);
    Cert.Validity.NotAfter.Choice := teUtcTime;
    Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now) + 365;
    Cert.Version := 3;
    Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints).ExtnValue.AsCe_BasicConstraints.CA := True;
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyCertSign];
    StreamSecII1.DefaultHashAlgorithm := HA;
    StreamSecII1.CreateKeyPair(Cert,OID,KeySize);
    StreamSecII1.SignCertificate(Cert,Cert.Data);
    X509TrustedCertificates1.AddCertificate(Cert.Data,True,Status);
    Assert(Status = crcOK);
  finally
    Cert.Free;
  end;
end;

procedure TdmDigSign.CreateECKeyPair;
var
  Cert: TCertificate;
  OID: string;
  Status: TCertStatusCode;
begin
  Cert := TCertificate.Create(nil,nil);
  try
    if NRKey then begin
      Cert.Subject.CommonName := 'ECNR Demo Key';
      OID := id_ecnr;
    end else begin
      Cert.Subject.CommonName := 'ECDSA Demo Key';
      OID := id_ecPublicKey;
    end;
    Cert.Validity.NotBefore.Choice := teUtcTime;
    Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now - X509TrustedCertificates1.HoursOffsetFromGMT/24);
    Cert.Validity.NotAfter.Choice := teUtcTime;
    Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now - X509TrustedCertificates1.HoursOffsetFromGMT/24) + 365;
    Cert.Version := 3;
    Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints).ExtnValue.AsCe_BasicConstraints.CA := True;
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyCertSign];
    StreamSecII1.DefaultHashAlgorithm := HA;
    StreamSecII1.CreateKeyPair(Cert,OID,0,Curve);
    StreamSecII1.SignCertificate(Cert,Cert.Data);
    X509TrustedCertificates1.AddCertificate(Cert.Data,True,Status);
    Assert(Status = crcOK);
  finally
    Cert.Free;
  end;
end;

procedure TdmDigSign.CreateIFKeyPair;
var
  Cert: TCertificate;
  Status: TCertStatusCode;
begin
  Cert := TCertificate.Create(nil,nil);
  try
    Cert.Validity.NotBefore.Choice := teUtcTime;
    Cert.Validity.NotBefore.AsUtcTime.AsDateTime := Trunc(Now);
    Cert.Validity.NotAfter.Choice := teUtcTime;
    Cert.Validity.NotAfter.AsUtcTime.AsDateTime := Trunc(Now) + 365;
    Cert.Version := 3;
    Cert.TbsCertificate.Extensions.KeyUsage := [digitalSignature,keyCertSign];
    Cert.TbsCertificate.Extensions.AddUniqueItem(eveIdCeBasicConstraints).ExtnValue.AsCe_BasicConstraints.CA := True;
    StreamSecII1.DefaultHashAlgorithm := HA;
    if PSSKey then begin
      Cert.Subject.CommonName := 'RSASSA-PSS Demo Key';
      StreamSecII1.CreateKeyPair(Cert,id_rsassa_pss,KeySize)
    end else begin
      Cert.Subject.CommonName := 'RSA Demo Key';
      StreamSecII1.CreateKeyPair(Cert,rsaEncryption,KeySize);
    end;
    StreamSecII1.SignCertificate(Cert,Cert.Data);
    X509TrustedCertificates1.AddCertificate(Cert.Data,True,Status);
    Assert(Status = crcOK);
  finally
    Cert.Free;
  end;
end;

function TdmDigSign.HasDLKeyPair(NRKey: Boolean): Boolean;
var
  Idx: Integer;
  Crt: PASN1Struct;
  Subject: TX501Name;
begin
  Idx := 0;
  repeat
    Result := X509TrustedCertificates1.FindCert([],
                                                [id_dsa],
                                                [digitalSignature],
                                                Idx,
                                                Crt);
    if Result then begin
      ExtractSubject(Crt^,Subject);
      Result := (NRKey and (Subject.commonName.Str = 'NR Demo Key')) xor
                ((not NRKey) and (Subject.commonName.Str = 'DSA Demo Key'));
      if Result then Break;
    end;
  until Crt = nil;
end;

function TdmDigSign.HasECKeyPair(NRKey: Boolean): Boolean;
var
  Idx: Integer;
  Crt: PASN1Struct;
  Subject: TX501Name;
begin
  Idx := 0;
  repeat
    Result := X509TrustedCertificates1.FindCert([],
                                                [id_ecPublicKey],
                                                [digitalSignature],
                                                Idx,
                                                Crt);
    if Result then begin
      ExtractSubject(Crt^,Subject);
      Result := (NRKey and (Subject.commonName.Str = 'ECNR Demo Key')) xor
                ((not NRKey) and (Subject.commonName.Str = 'ECDSA Demo Key'));
      if Result then Break;
    end;
  until Crt = nil;
end;

function TdmDigSign.HasIFKeyPair(PSSKey: Boolean): Boolean;
var
  Idx: Integer;
  Crt: PASN1Struct;
begin
  Idx := 0;
  if PSSKey then
    repeat
      Result := X509TrustedCertificates1.FindCert([],
                                                  [id_rsaes_oaep,id_rsassa_pss],
                                                  [digitalSignature],
                                                  Idx,
                                                  Crt);
      if Result then Break;
    until Crt = nil
  else
    repeat
      Result := X509TrustedCertificates1.FindCert([],
                                                  [rsaEncryption],
                                                  [digitalSignature],
                                                  Idx,
                                                  Crt);
      if Result then Break;
    until Crt = nil;
end;

function TdmDigSign.SignMsg(const Msg: string; HA: THashAlgorithm): string;
var
  Crt: TASN1Struct;
  PKI: string;
  SignLen: Integer;
begin
  // Sign Msg using the key that belongs to the last certificate in the store.
  Crt := X509TrustedCertificates1.Certs[X509TrustedCertificates1.Count - 1];
  PKI := ExtractSubjectKeyIdentifier(Crt);
  // 512 bytes = 4096 bits. It's large enough for all keys in this demo.
  SetLength(Result,512);
  SignLen := Length(Result);
  Assert(StreamSecII1.SignBuf(Pointer(Msg)^,Length(Msg),
                              Pointer(PKI),Length(PKI),
                              HA,
                              Pointer(Result)^,SignLen));
  SetLength(Result,SignLen);
end;

function TdmDigSign.VerifySign(const Msg, Sign: string;
  HA: THashAlgorithm): Boolean;
var
  Crt: TASN1Struct;           
  Subject: TX501Name;
  DLKey: TDLPublicKey;
  ECKey: TECPublicKey;
  IFKey: TIFPublicKey;
begin
  Crt := X509TrustedCertificates1.Certs[X509TrustedCertificates1.Count - 1];
  // Normally, you would use the subject public key algorithm identifier to
  // identify the public key algorithm. This code demonstrates the use of
  // digital signatures and *not* an entirely proper use of X.509 certificates.
  ExtractSubject(Crt,Subject);
  if Subject.commonName.Str = 'DSA Demo Key' then begin
    FillChar(DLKey,SizeOf(DLKey),0);
    ExtractSubjectDSAPublicKey(Crt,DLKey);
    Result := DLSSASignatureVerification(DLKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         True,HA);
    DisposeDLPublicKey(DLKey);
  end else if Subject.commonName.Str = 'NR Demo Key' then begin
    FillChar(DLKey,SizeOf(DLKey),0);
    ExtractSubjectDSAPublicKey(Crt,DLKey);
    DLKey.SignScheme := dlssaNR;
    Result := DLSSASignatureVerification(DLKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         True,HA);
    DisposeDLPublicKey(DLKey);
  end else if Subject.commonName.Str = 'ECDSA Demo Key' then begin
    FillChar(ECKey,SizeOf(ECKey),0);
    ExtractSubjectECPublicKey(Crt,ECKey);
    Result := ECSSASignatureVerification(ECKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         True,HA);
    DisposeECPublicKey(ECKey);
  end else if Subject.commonName.Str = 'ECNR Demo Key' then begin
    FillChar(ECKey,SizeOf(ECKey),0);
    ExtractSubjectECPublicKey(Crt,ECKey);
    ECKey.SignScheme := ecssaNR;
    Result := ECSSASignatureVerification(ECKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         True,HA);
    DisposeECPublicKey(ECKey);
  end else if Subject.commonName.Str = 'RSA Demo Key' then begin
    FillChar(IFKey,SizeOf(IFKey),0);
    ExtractSubjectRSAPublicKey(Crt,IFKey);
    Result := IFSSASignatureVerification(IFKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         HA,haSHA1,seEMSA3);
    DisposeIFPublicKey(IFKey);
  end else if Subject.commonName.Str = 'RSASSA-PSS Demo Key' then begin
    FillChar(IFKey,SizeOf(IFKey),0);
    ExtractSubjectRSAPublicKey(Crt,IFKey,True);
    Result := IFSSASignatureVerification(IFKey,
                                         Pointer(Msg)^,Length(Msg),
                                         Pointer(Sign)^,Length(Sign),
                                         HA,haSHA1,seEMSA4);
    DisposeIFPublicKey(IFKey);
  end else
    Result := False;
end;

procedure TdmDigSign.StreamSecII1AdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetLength(3);
  Password.SetKeyStrAt('abc',0);
end;

procedure TdmDigSign.StreamSecII1Password(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetLength(3);
  Password.SetKeyStrAt('def',0);
end;

end.
