unit QSignCodeDM;

interface

uses
  SysUtils,
  {$IFDEF WIN32}
  Forms,
  {$ENDIF}
  Classes, SecUtils, MpCms, MpSignCode, MpX509, StreamSecII, SecComp, Asn1;

type
  TdmQSignCode = class(TDataModule)
    PrivateKeyRing: TStreamSecII;
    RootCerts: TX509TrustedCertificates;
    MyCerts: TX509TrustedCertificates;
    SignCode: TMPSignCode;
    CollectedCerts: TX509TrustedCertificates;
    procedure PrivateKeyRingAdminPassword(Sender: TObject;
      Password: ISecretKey);
    procedure PrivateKeyRingPassword(Sender: TObject;
      Password: ISecretKey);
    procedure DataModuleCreate(Sender: TObject);
    procedure CollectedCertsCertNotTrusted(Sender: TObject;
      Cert: TASN1Struct; var ExplicitTrust: Boolean);
  private
    FURL: string;
    FPrivateKeyRingFileName: WideString;
    FRootCertificatesFileName: WideString;
    FExeFileName: WideString;
    FMyCertificatesFileName: WideString;
    FDescription: WideString;
    FCounterSign: Boolean;
    FIncludeSigningTime: Boolean;
    FStatementType: Integer;
    procedure SetDescription(const Value: WideString);
    procedure SetExeFileName(const Value: WideString);
    procedure SetMyCertificatesFileName(const Value: WideString);
    procedure SetPrivateKeyRingFileName(const Value: WideString);
    procedure SetRootCertificatesFileName(const Value: WideString);
    procedure SetURL(const Value: string);
    procedure SetCounterSign(const Value: Boolean);
    procedure SetIncludeSigningTime(const Value: Boolean);
    procedure SetStatementType(const Value: Integer);
  public
    function PerformSignCode: Boolean;
    property ExeFileName: WideString read FExeFileName write SetExeFileName;
    property CounterSign: Boolean read FCounterSign write SetCounterSign;
    property IncludeSigningTime: Boolean read FIncludeSigningTime write SetIncludeSigningTime;
    property StatementType: Integer read FStatementType write SetStatementType;
    property Description: WideString read FDescription write SetDescription;
    property URL: string read FURL write SetURL;
    property PrivateKeyRingFileName: WideString read FPrivateKeyRingFileName write SetPrivateKeyRingFileName;
    property MyCertificatesFileName: WideString read FMyCertificatesFileName write SetMyCertificatesFileName;
    property RootCertificatesFileName: WideString read FRootCertificatesFileName write SetRootCertificatesFileName;
  end;

var
  dmQSignCode: TdmQSignCode;

implementation

uses                           
  {$IFDEF LINUX}
  QForms,
  {$ENDIF}
  PasswordMain, IniFiles;

{$IFDEF LINUX}
{$R *.xfm}
{$ENDIF}
{$IFDEF WIN32}
{$R *.dfm}
{$ENDIF}


{ TdmQSignCode }

function TdmQSignCode.PerformSignCode: Boolean;
var
  IniFile: TIniFile;
  FN: string;
begin
  if FCounterSign then
    Result := SignCode.CounterSignCode(FExeFileName)
  else
    Result := SignCode.SignCode(FExeFileName);
  if not Result then begin
    if SignCode.IllegalFormat then
      raise Exception.Create(Format('File %s is not a valid Win32 PE image',[FExeFileName]));
    if SignCode.CertificateNotFound then
      raise Exception.Create('Could not find a private key and a matching code signing certificate');
    raise Exception.Create(Format('Could not sign %s',[FExeFileName]));
  end else begin
    FN := ChangeFileExt(Application.ExeName,'.ini');
    IniFile := TIniFile.Create(FN);
    try
      IniFile.WriteString('Info','URL',FURL);
      IniFile.WriteString('Keys','MyCerts',FMyCertificatesFileName);
      IniFile.WriteString('Keys','RootCerts',FRootCertificatesFileName);
      IniFile.WriteString('Keys','PKR',FPrivateKeyRingFileName);
    finally
      IniFile.Free;
    end;
  end;
end;

procedure TdmQSignCode.SetDescription(const Value: WideString);
begin
  FDescription := Value;
  SignCode.Description := Value;
end;

procedure TdmQSignCode.SetExeFileName(const Value: WideString);
begin
  FExeFileName := Value;
end;

procedure TdmQSignCode.SetMyCertificatesFileName(const Value: WideString);
var
  Ext: string;
  FS: TFileStream;
  Cert: TASN1Struct;
  StatusCode: TCertStatusCode;
  Status: TCertStatusCodes;
begin                              
//  if FMyCertificatesFileName = Value then Exit;

  FMyCertificatesFileName := Value;
  Ext := ExtractFileExt(Value);
  if Ext = '.scl' then begin
    MyCerts.LoadFromFile(Value);
  end else if Ext = '.cer' then begin
    FS := TFileStream.Create(Value,fmOpenRead);
    try
      Cert := TASN1Struct.Create;
      try
        Cert.LoadFromStream(FS,fmtDER);
        MyCerts.AddCertificate(Cert,True,StatusCode);
      finally
        Cert.Free;
      end;
    finally
      FS.Free;
    end;
  end else if Ext = '.p7b' then begin
    MyCerts.ImportFromCMSFile(Value,True,Status);
  end else
    raise Exception.Create('Unknown file format: "*' + Ext + '"');
end;

procedure TdmQSignCode.SetPrivateKeyRingFileName(const Value: WideString);
var
  FS: TFileStream;
  PW: ISecretKey;
begin                              
//  if FPrivateKeyRingFileName = Value then Exit;

  FPrivateKeyRingFileName := Value;
  FS := TFileStream.Create(Value,fmOpenRead);
  try
    PW := TSecretKey.Create('');
    PrivateKeyRingPassword(nil,PW);
    if not PrivateKeyRing.LoadPrivateKeyRingFromStream(FS,PW) then
      raise Exception.Create('Unable to load private key ring from file ' + Value);
  finally
    FS.Free;
  end;
end;

procedure TdmQSignCode.SetRootCertificatesFileName(
  const Value: WideString);
var
  Ext: string;
  FS: TFileStream;
  Cert: TASN1Struct;
  StatusCode: TCertStatusCode;
  Status: TCertStatusCodes;
begin
  if not FileExists(Value) then Exit;
//  if FRootCertificatesFileName = Value then Exit;

  FRootCertificatesFileName := Value;
  Ext := ExtractFileExt(Value);
  if Ext = '.scl' then begin
    RootCerts.LoadFromFile(Value);
  end else if Ext = '.cer' then begin
    FS := TFileStream.Create(Value,fmOpenRead);
    try
      Cert := TASN1Struct.Create;
      try
        Cert.LoadFromStream(FS,fmtDER);
        RootCerts.AddCertificate(Cert,True,StatusCode);
      finally
        Cert.Free;
      end;
    finally
      FS.Free;
    end;
  end else if Ext = '.p7b' then begin
    RootCerts.ImportFromCMSFile(Value,True,Status);
  end else
    raise Exception.Create('Unknown file format: "*' + Ext + '"');
end;

procedure TdmQSignCode.SetURL(const Value: string);
begin
  FURL := Value;
  SignCode.URL := Value;
end;

procedure TdmQSignCode.PrivateKeyRingAdminPassword(Sender: TObject;
  Password: ISecretKey);
begin
  PasswordDlg('Private Key Ring Admin',Password);
end;

procedure TdmQSignCode.PrivateKeyRingPassword(Sender: TObject;
  Password: ISecretKey);
begin
  PasswordDlg('Private Key Ring',Password);
end;

procedure TdmQSignCode.DataModuleCreate(Sender: TObject);
var
  FN, S: string;
  IniFile: TIniFile;
begin
  FN := ChangeFileExt(Application.ExeName,'.ini');
  IniFile := TIniFile.Create(FN);
  try
    S := IniFile.ReadString('Info','URL','');
    if S <> '' then
      URL := S;
    S := IniFile.ReadString('Keys','MyCerts','');
    if S <> '' then
      FMyCertificatesFileName := S;
    S := IniFile.ReadString('Keys','RootCerts','');
    if S <> '' then
      FRootCertificatesFileName := S;
    S := IniFile.ReadString('Keys','PKR','');
    if S <> '' then
      FPrivateKeyRingFileName := S;
  finally
    IniFile.Free;
  end;
end;

procedure TdmQSignCode.SetCounterSign(const Value: Boolean);
begin
  FCounterSign := Value;
end;

procedure TdmQSignCode.CollectedCertsCertNotTrusted(Sender: TObject;
  Cert: TASN1Struct; var ExplicitTrust: Boolean);
begin
  ExplicitTrust := True;
end;

procedure TdmQSignCode.SetIncludeSigningTime(const Value: Boolean);
begin
  FIncludeSigningTime := Value;
  SignCode.IncludeSigningTime := Value;
end;

procedure TdmQSignCode.SetStatementType(const Value: Integer);
begin
  FStatementType := Value;
  SignCode.StatementType := TStatementType(Value);
end;

end.
