unit TLSChatServerDM;

interface

uses
  SysUtils, Classes, Forms,
  Windows, SyncObjs, MpYarrow, MpX509, StreamSecII,
  ScktComp, Tls, TlsClass, SecUtils, TlsInternalServer, TlsSocket,
  ResourceFile;

type
  TdmTLSChatServer = class(TDataModule)
    TLSServerSocket1: TTLSServerSocket;
    PrivateKeyRing: TResourceFile;
    RootCert: TResourceFile;
    ServerCert: TResourceFile;
    TLSInternalServer1: TTLSInternalServer;
    procedure ApplicationEvents1Message(var Msg: tagMSG;
      var Handled: Boolean);
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure ClientSocket1Connect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure StreamSecII1TLSIncomingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure StreamSecII1TLSOutgoingAlert(Sender: TObject;
      Client: TCustomTLS_ContentLayer; var Fatal: Boolean;
      AlertCode: Integer);
    procedure StreamSecII1TLSChangeCipherSpec(Sender: TObject;
      Client: TCustomTLS_ContentLayer);
    procedure TLSInternalServer1Password(Sender: TObject;
      Password: ISecretKey);
    procedure TLSServerSocket1NewConnect(Sender: TObject;
      Socket: TCustomTLSSocket);
  private
    FLog: TStrings;
    FLogLock: TCriticalSection;
    procedure Broadcast(const Str: string);
    procedure LogEntry(const Entry: string);
  public
    procedure Start(const Port: string);
    procedure Stop;
    procedure AppendLog(Dst: TStrings);
    function HasPendLogEntries: Boolean;
  end;

var
  dmTLSChatServer: TdmTLSChatServer;

implementation

uses
  ReadStrm, Pkix, Asn1, X509Base, TLSChatServerMain;

{$R *.dfm}

procedure TdmTLSChatServer.ApplicationEvents1Message(var Msg: tagMSG;
  var Handled: Boolean);
begin
  // This is an acceptable way to collect entropy for random number generation:
  TLSInternalServer1.AccumulateEntropy;
end;

procedure TdmTLSChatServer.Start(const Port: string);
begin
  TLSInternalServer1.Options.SignatureRSA := prPrefer;
  TLSInternalServer1.Options.KeyAgreementRSA := prAllowed;
  TLSInternalServer1.Options.KeyAgreementDHE := prPrefer;
  TLSInternalServer1.TLSSetupServer;
  TLSServerSocket1.Port := StrToInt(Port);
  TLSServerSocket1.Active := True;
  Application.OnMessage := ApplicationEvents1Message;
end;

procedure TdmTLSChatServer.DataModuleCreate(Sender: TObject);
begin
  FLogLock := TCriticalSection.Create;
  FLog := TStringList.Create;
  Application.OnMessage := ApplicationEvents1Message;
end;

procedure TdmTLSChatServer.DataModuleDestroy(Sender: TObject);
begin
  FLog.Free;
  FLogLock.Free;
end;

procedure TdmTLSChatServer.LogEntry(const Entry: string);
begin
  FLogLock.Acquire;
  try
    FLog.Add(DateTimeToStr(Now) + #13#10 + Entry + #13#10);
  finally
    FLogLock.Release;
  end;
end;

procedure TdmTLSChatServer.ClientSocket1Connect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  LogEntry('Client connecting: ' + Socket.RemoteHost + #13#10);
end;

procedure TdmTLSChatServer.StreamSecII1TLSIncomingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Incoming close notification'#13#10
  else if Fatal then
    Msg := 'Incoming fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Incoming warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The remote host has terminated the connection.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations. Please notify StreamSec if you get this message.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc.';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;

  LogEntry(Msg + #13#10);
end;

procedure TdmTLSChatServer.StreamSecII1TLSOutgoingAlert(Sender: TObject;
  Client: TCustomTLS_ContentLayer; var Fatal: Boolean; AlertCode: Integer);
var
  Msg: string;
begin
  if AlertCode = 0 then
    Msg := 'Outgoing close notification'#13#10
  else if Fatal then
    Msg := 'Outgoing fatal alert ' + IntToStr(AlertCode) + #13#10
  else
    Msg := 'Outgoing warning ' + IntToStr(AlertCode) + #13#10;
  case AlertCode and $FF of
    0:  Msg := Msg + 'close_notify: The connection is about to be terminated.';
    10: Msg := Msg + 'unexpected_message:'#13#10 +
       'An inappropriate message was received. This alert is always fatal'#13#10 +
       'and should never be observed in communication between proper'#13#10 +
       'implementations.';
    20: Msg := Msg + 'bad_record_mac:'#13#10 +
       'This alert is returned if a record is received with an incorrect'#13#10 +
       'MAC. This alert also SHOULD be returned if a TLSCiphertext'#13#10 +
       'decrypted in an invalid way: either it wasn''t an even multiple of'#13#10 +
       'the block length, or its padding values, when checked, weren''t'#13#10 +
       'correct. This message is always fatal.';
    21: Msg := Msg + 'decryption_failed:'#13#10 +
       'This alert MAY be returned if a TLSCiphertext decrypted in an'#13#10 +
       'invalid way: either it wasn''t an even multiple of the block'#13#10 +
       'length, or its padding values, when checked, weren''t correct.'#13#10 +
       'This message is always fatal.'#13#10 +
       #13#10 +
       'NB: Differentiating between bad_record_mac and decryption_failed'#13#10 +
       'alerts may permit certain attacks against CBC mode as used in'#13#10 +
       'TLS. It is preferable to uniformly use the bad_record_mac alert'#13#10 +
       'to hide the specific type of the error.';
    22: Msg := Msg + 'record_overflow:'#13#10 +
       'A TLSCiphertext record was received which had a length more than'#13#10 +
       '2^14+2048 bytes, or a record decrypted to a TLSCompressed record'#13#10 +
       'with more than 2^14+1024 bytes. This message is always fatal.';
    30: Msg := Msg + 'decompression_failure:'#13#10 +
       'The decompression function received improper input (e.g. data'#13#10 +
       'that would expand to excessive length). This message is always'#13#10 +
       'fatal.';
    40: Msg := Msg + 'handshake_failure:'#13#10 +
       'Reception of a handshake_failure alert message indicates that the'#13#10 +
       'sender was unable to negotiate an acceptable set of security'#13#10 +
       'parameters given the options available. This is a fatal error.'#13#10 +
       'IMPORTANT: This error is likely to occur with this demo if a client'#13#10 +
       'tries to connect before the system parameters for the DHE key exchange'#13#10 +
       'have been generated. This might take a couple of seconds after'#13#10 +
       'you pressed the Start button.';
    41: Msg := Msg + 'no_certificate_RESERVED:'#13#10 +
       'This alert was used in SSLv3 but not in TLS. It should not be'#13#10 +
       'sent by compliant implementations.';
    42: Msg := Msg + 'bad_certificate:'#13#10 +
       'A certificate was corrupt, contained signatures that did not'#13#10 +
       'verify correctly, etc.';
    43: Msg := Msg + 'unsupported_certificate:'#13#10 +
       'A certificate was of an unsupported type.';
    44: Msg := Msg + 'certificate_revoked:'#13#10 +
       'A certificate was revoked by its signer.';
    45: Msg := Msg + 'certificate_expired:'#13#10 +
       'A certificate has expired or is not currently valid.';
    46: Msg := Msg + 'certificate_unknown:'#13#10 +
       'Some other (unspecified) issue arose in processing the'#13#10 +
       'certificate, rendering it unacceptable.';
    47: Msg := Msg + 'illegal_parameter:'#13#10 +
       'A field in the handshake was out of range or inconsistent with'#13#10 +
       'other fields. This is always fatal.';
    48: Msg := Msg + 'unknown_ca:'#13#10 +
       'A valid certificate chain or partial chain was received, but the'#13#10 +
       'certificate was not accepted because the CA certificate could not'#13#10 +
       'be located or couldn''t be matched with a known, trusted CA.  This'#13#10 +
       'message is always fatal.';
    49: Msg := Msg + 'access_denied:'#13#10 +
       'A valid certificate was received, but when access control was'#13#10 +
       'applied, the sender decided not to proceed with negotiation.'#13#10 +
       'This message is always fatal.';
    50: Msg := Msg + 'decode_error:'#13#10 +
       'A message could not be decoded because some field was out of the'#13#10 +
       'specified range or the length of the message was incorrect. This'#13#10 +
       'message is always fatal.';
    51: Msg := Msg + 'decrypt_error:'#13#10 +
       'A handshake cryptographic operation failed, including being'#13#10 +
       'unable to correctly verify a signature, decrypt a key exchange,'#13#10 +
       'or validate a finished message.';
    60: Msg := Msg + 'export_restriction:'#13#10 +
       'A negotiation not in compliance with export restrictions was'#13#10 +
       'detected; for example, attempting to transfer a 1024 bit'#13#10 +
       'ephemeral RSA key for the RSA_EXPORT handshake method. This'#13#10 +
       'message is always fatal.';
    70: Msg := Msg + 'protocol_version:'#13#10 +
       'The protocol version the client has attempted to negotiate is'#13#10 +
       'recognized, but not supported. (For example, old protocol'#13#10 +
       'versions might be avoided for security reasons). This message is'#13#10 +
       'always fatal.';
    71: Msg := Msg + 'insufficient_security:'#13#10 +
       'Returned instead of handshake_failure when a negotiation has'#13#10 +
       'failed specifically because the server requires ciphers more'#13#10 +
       'secure than those supported by the client. This message is always'#13#10 +
       'fatal.';
    80: Msg := Msg + 'internal_error:'#13#10 +
       'An internal error unrelated to the peer or the correctness of the'#13#10 +
       'protocol makes it impossible to continue (such as a memory'#13#10 +
       'allocation failure). This message is always fatal.';
    90: Msg := Msg + 'user_canceled:'#13#10 +
       'This handshake is being canceled for some reason unrelated to a'#13#10 +
       'protocol failure. If the user cancels an operation after the'#13#10 +
       'handshake is complete, just closing the connection by sending a'#13#10 +
       'close_notify is more appropriate. This alert should be followed'#13#10 +
       'by a close_notify. This message is generally a warning.';
    100: Msg := Msg + 'no_renegotiation:'#13#10 +
       'Sent by the client in response to a hello request or by the'#13#10 +
       'server in response to a client hello after initial handshaking.'#13#10 +
       'Either of these would normally lead to renegotiation; when that'#13#10 +
       'is not appropriate, the recipient should respond with this alert'#13#10 +
       'at that point, the original requester can decide whether to'#13#10 +
       'proceed with the connection. One case where this would be'#13#10 +
       'appropriate would be where a server has spawned a process to'#13#10 +
       'satisfy a request; the process might receive security parameters'#13#10 +
       '(key length, authentication, etc.) at startup and it might be'#13#10 +
       'difficult to communicate changes to these parameters after that'#13#10 +
       'point. This message is always a warning.';
  else
    Msg := Msg + 'Unknown alert code: No description available.';
  end;

  LogEntry(Msg + #13#10);
end;

procedure TdmTLSChatServer.StreamSecII1TLSChangeCipherSpec(
  Sender: TObject; Client: TCustomTLS_ContentLayer);
var
  KS: Integer;
  PK: TX509PublicKey;
  Alg: TX509AlgorithmIdentifier;
  XName: TX501Name;
  Msg: string;
begin
  Msg := 'Connection is encrypted:'#13#10;
  if Client.ServerCertificate = nil then
    Msg := Msg + 'Warning: Anonymous connection - vulnerable to man in the middle attacks.'#13#10
  else begin
    KS := 0;
    ExtractSubjectPublicKeySize(Client.ServerCertificate^,KS);
    Msg := Msg + '  Server certificate:'#13#10 +
      '    Public Key Size (normalized ): ' +
      IntToStr(KS) + #13#10;
    PK := nil;
    try
      if ExtractSubjectPublicKey(Client.ServerCertificate^,PK) = E_OK then
        Msg := Msg + '    Public Key Algorithm: ' +
          GetObjectName(PK.Algorithm.Algorithm) + #13#10;
    finally
      PK.Free;
    end;
    Alg := nil;
    try
      if ExtractSignatureAlgorithmCertOrCRL(Client.ServerCertificate^,Alg) = E_OK then
        Msg := Msg + '    Signature Algorithm: ' +
          GetObjectName(Alg.Algorithm) + #13#10;
    finally
      Alg.Free;
    end;

    if Assigned(Client.ClientCertificate) then begin
      Msg := Msg + '  Client certificate:'#13#10;
      ExtractSubject(Client.ClientCertificate^,XName);
      Msg := Msg + '    Client Subject Name: ' + XName.commonName.Str + #13#10;
      XName.commonName.Str := '';
      ExtractIssuer(Client.ClientCertificate^,XName);
      Msg := Msg + '    Client Certificate Issuer Name: ' + XName.commonName.Str + #13#10;
    end;
  end;

  Msg := Msg + '  Symmetric encryption key size: ' +
    IntToStr(Client.Context^.key_size*8) + ' bits'#13#10;
  case Client.Context^.bulk_cipher_algorithm of
    bcaRC4,
    bcaRC4_40,
    bcaRC4_56: Msg := Msg + '  Symmetric encryption algorithm: RC4'#13#10;
    bcaDES,
    bcaDES40,
    bcaDES56: Msg := Msg + '  Symmetric encryption algorithm: DES'#13#10;
    bca3DES:  Msg := Msg + '  Symmetric encryption algorithm: TripleDES'#13#10;
    bcaAES128,
    bcaAES256:Msg := Msg + '  Symmetric encryption algorithm: AES'#13#10;
  end;
  case Client.Context^.mac_algorithm of
    haMD5:  Msg := Msg + '  MAC algorithm: MD5'#13#10;
    haSHA1: Msg := Msg + '  MAC algorithm: SHA1'#13#10;
  end;

  LogEntry(Msg + #13#10);
end;

procedure TdmTLSChatServer.Broadcast(const Str: string);
var
  I: Integer;
  Socket: TCustomTLSSocket;
begin
  LogEntry('Data:'#13#10 + Str);
  for I := 0 to TLSServerSocket1.TLSSocketCount - 1 do begin
    Socket := TLSServerSocket1.TLSSockets[I];
    Socket.Lock;
    try
      Socket.SendBuf(Pointer(Str)^,Length(Str));
    finally
      Socket.Unlock;
    end;
  end;
end;

procedure TdmTLSChatServer.TLSInternalServer1Password(Sender: TObject;
  Password: ISecretKey);
begin
  Password.SetLength(3);
  Password.SetKeyStrAt('abc',0);
end;

procedure TdmTLSChatServer.TLSServerSocket1NewConnect(Sender: TObject;
  Socket: TCustomTLSSocket);
var
  MsgLen: Integer;
  Msg: string;
begin
  LogEntry('Client connected: ' + Socket.RemoteHost + #13#10);
  try
    while Socket.Connected do begin
      MsgLen := Socket.ReceiveLength;
      if MsgLen > 0 then begin
        SetLength(Msg,MsgLen);
        Socket.ReceiveBuf(Pointer(Msg)^,MsgLen);
        LogEntry('Incoming message' + #13#10);
        Broadcast(Msg);
      end;
    end;
  except
    on E: Exception do
      LogEntry('Client exception: ' + Socket.RemoteHost + #13#10 +
               '  Error message: ' + E.Message + #13#10);
  end;
  LogEntry('Client disconnected: ' + Socket.RemoteHost + #13#10);
end;

procedure TdmTLSChatServer.AppendLog(Dst: TStrings);
begin
  FLogLock.Acquire;
  try
    Dst.AddStrings(FLog);
    FLog.Clear;
  finally
    FLogLock.Release;
  end;
end;

function TdmTLSChatServer.HasPendLogEntries: Boolean;
begin
  FLogLock.Acquire;
  try
    Result := FLog.Count > 0;
  finally
    FLogLock.Release;
  end;
end;

procedure TdmTLSChatServer.Stop;
begin
  TLSServerSocket1.Active := False;
  Application.OnMessage := nil;
  TLSInternalServer1.StopEphemeralKeyGen;
end;

end.
